/*
 * BMP (24Bit) Converter for S65 Display
 *
 * by Juergen Eckert
 */

#include <stdio.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

struct bmp_header 				//struct alignment tested on 32bit arch
{
	unsigned char ID_lo;
	unsigned char ID_hi;
	//2 dummybytes
	unsigned int filesize;
	unsigned int nix;
	unsigned int posofpic;
	unsigned int lengthofheader;
	unsigned int width;
	unsigned int height;
	unsigned char layers_lo;
	unsigned char layers_hi;
	unsigned char depth_lo;
	unsigned char depth_hi;
	unsigned int compression;
	unsigned int size;
	unsigned int h_dpi;
	unsigned int v_dpi;
	unsigned int numusedcolors;
	unsigned int numimpcolors;
};

int main(int argc, char *argv[])
{
	FILE *fd_in;
	int i, j, inp, compress = 0;
	unsigned char header[56];
	struct bmp_header *bmp_head;
	unsigned short *buffer;
	unsigned char *buffer8;
	int targetcolordepth;
	int srcsize=0, targetsize=0, compressedtargetsize=0;

	if(argc != 4)
	{
		fprintf(stderr, "BMP (24-Bit) to H-File converter\n");
		fprintf(stderr, "Usage: %s inputfile picturename 16|15|12|10|8\n", argv[0]);
		exit(1);
	}

	if((fd_in=fopen(argv[1],"r"))==0)
	{
		fprintf(stderr,"Could not open inputfile: %s\n",argv[1]);
        	exit(-1);
	}
		
	//Header
	fprintf(stderr, "Reading Header...\n");
	for(i=0; i<56; i++)
	{
		inp = fgetc(fd_in);
		if(inp == EOF)
		{
			fprintf(stderr,"Unexpected EOF\n");
			return(0);
		}
		if(i==2)
			i+=2; 
		header[i] = (unsigned char) inp;
	}
	bmp_head = (struct bmp_header *) header;
	if(bmp_head->ID_lo != 0x42 || bmp_head->ID_hi != 0x4D)
	{
		fprintf(stderr,"BM");
		return(0);
	}
	if(bmp_head->width * bmp_head->height > 132*176)
	{
		fprintf(stderr,"File to big\n");
		return(0);	
	}
	if(bmp_head->depth_lo != 24 || bmp_head->depth_hi != 0)
	{
		fprintf(stderr,"Depth unsupported\n");
		return(0);	
	}
	
	fprintf(stderr, "Width: %u\nHeight: %u\nSourcedepth: %u\n", bmp_head->width, bmp_head->height, bmp_head->depth_lo);
	
	targetcolordepth = atoi(argv[3]);
	fprintf(stderr,"Targetdepth: %i\n", targetcolordepth);

	if(targetcolordepth == 15 ||targetcolordepth == 12 || targetcolordepth == 10)
	{
		compress = 1;
	}
	else
	{
		compress = 0;
	}


	switch(targetcolordepth)
	{
	case(8):
		printf("prog_uint8_t %s[] = {\n", argv[2]);
		printf("0x08, 0x00, ");
		printf("0x%02x, ", (0xFF & bmp_head->width));
		printf("0x%02x,\n", (0xFF & bmp_head->height));
		break;
	case(10):
		compress = 1;
		printf("prog_uint16_t %s[] = {\n", argv[2]);
		printf("0x0110, ");
		printf("0x%04x,\n", (0xFF00 & (bmp_head->height<<8))|(0xFF & bmp_head->width));
		break;
	case(12):
		compress = 1;
		printf("prog_uint16_t %s[] = {\n", argv[2]);
		printf("0x0112, ");
		printf("0x%04x,\n", (0xFF00 & (bmp_head->height<<8))|(0xFF & bmp_head->width));
		break;
	case(15):
		compress = 1;
		printf("prog_uint16_t %s[] = {\n", argv[2]);
		printf("0x0115, ");
		printf("0x%04x,\n", (0xFF00 & (bmp_head->height<<8))|(0xFF & bmp_head->width));
		break;
	default:
		fprintf(stderr,"unsupported color depth -> setting to 16bit\n");
		targetcolordepth = 16;
	case(16):
		printf("prog_uint16_t %s[] = {\n", argv[2]);
		printf("0x0016, ");
		printf("0x%04x,\n", (0xFF00 & (bmp_head->height<<8))|(0xFF & bmp_head->width));
		break;
	}		


	if(targetcolordepth == 16 || targetcolordepth == 15 || targetcolordepth == 12 || targetcolordepth == 10)
	{
		buffer = malloc(sizeof(unsigned short) * (bmp_head->width*bmp_head->height));
		if(buffer == NULL)
		{
			fprintf(stderr, "malloc\n");
			return(0);
		}
	}
	else if(targetcolordepth == 8)
	{
		buffer8 = malloc(sizeof(unsigned short) * (bmp_head->width*bmp_head->height));
		if(buffer8 == NULL)
		{
			fprintf(stderr, "malloc\n");
			return(0);
		}
	}

	for(i=0; i<bmp_head->height; i++)
	{
		unsigned int aligment = 0;
		for(j=0; j<bmp_head->width; j++)
		{
			unsigned int color;
			unsigned char r,g,b,color8;
			inp = fgetc(fd_in);
			if(inp == EOF)
			{
				fprintf(stderr,"Unexpected EOF\n");
				return(0);
			}
			b = (unsigned char) inp;
			inp = fgetc(fd_in);
			aligment++;
			if(inp == EOF)
			{
				fprintf(stderr,"Unexpected EOF\n");
				return(0);
			}
			g = (unsigned char) inp;
			inp = fgetc(fd_in);
			aligment++;
			if(inp == EOF)
			{
				fprintf(stderr,"Unexpected EOF\n");
				return(0);
			}
			r = (unsigned char) inp;
			aligment++;
	
		//	fprintf(stderr,"Color: %X - %X - %X\n", r, b, g);

			if(targetcolordepth == 16 || targetcolordepth == 15 || targetcolordepth == 12 || targetcolordepth == 10)
			{
				color = 0xFFFF & (((r>>3)<<11) | ((g>>2)<<5) | ((b>>3)<<0)) ;
				buffer[(bmp_head->height-1-i)*bmp_head->width +j] = color;
				targetsize += 2;
			}
			else if(targetcolordepth == 8)
			{
				color8 = 0xFF & (((r>>5)<<5) | ((g>>5)<<2) | ((b>>6)<<0)) ;
				buffer8[(bmp_head->height-1-i)*bmp_head->width +j] = color8;
				targetsize += 1;
			}
			srcsize += 3;
		//	fprintf(stderr,"%i - %X - %X\n", (bmp_head->height-1-i)*bmp_head->width +j, color, buffer[(bmp_head->height-1-i)*bmp_head->width +j]);
		}
		while(aligment%4)
		{
			fgetc(fd_in);
			aligment++;
		}
	}
	fprintf(stderr, "Source Size: %u Bytes\n", srcsize);
	fprintf(stderr, "Target Size: %u Bytes\n", targetsize);
	if(compress)
	{
		unsigned int column = 0;
		fprintf(stderr, "RLE Kompression with %ibit colors\n", targetcolordepth);
		for(i=0; i<(bmp_head->width*bmp_head->height); i++)
		{
			switch(targetcolordepth)	//org: 5-6-5
			{
				case(10):		//3-4-3
					buffer[i] = ( ((buffer[i]&0xE000)>>6)|((buffer[i]&0x0780)>>4)|((buffer[i]&0x001C)>>2) );
					break;
				case(12):		//4-4-4
					buffer[i] = 0x7FFF & ( ((buffer[i]&0xF000)>>4)|((buffer[i]&0x0780)>>3)|((buffer[i]&0x001E)>>1) );
					break;
				case(15):		//5-5-5
					buffer[i] = 0x7FFF & (((buffer[i] & 0xFFC0 )>>1) | (buffer[i] & 0x001F));
					break;
			}
		}

		for(i=0; i<(bmp_head->width*bmp_head->height); i++)
		{
			if(targetcolordepth == 15)
			{
				unsigned short amount;
				amount=1;
				while((buffer[i+amount]==buffer[i]) && (amount<0x7FFF) && ((amount+i)<(bmp_head->width*bmp_head->height)))
				{
					amount++;
				}
				if(amount>2)
				{
					printf("0x%04x, ", 0x8000 + amount);
					if(!((column+1)%12))
						printf("\n");
					column++;
				}
				printf("0x%04x, ", buffer[i]);
				if(!((column+1)%12))
					printf("\n");
				column++;
				if(amount>2)
				{
					i+=(amount-1);
					compressedtargetsize+=2;
				}
				compressedtargetsize+=2;
			}
			else if(targetcolordepth == 12)
			{
				unsigned short amount;
				amount=1;
				while((buffer[i+amount]==buffer[i]) && (amount<0x10) && ((amount+i)<(bmp_head->width*bmp_head->height)))
				{
					amount++;
				}
				printf("0x%04x, ", ((amount-1)<<12) + buffer[i]);
				if(!((column+1)%12))
					printf("\n");
				column++;
				i+=(amount-1);
				compressedtargetsize+=2;
			}
			else if(targetcolordepth == 10)
			{
				unsigned short amount;
				amount=1;
				while((buffer[i+amount]==buffer[i]) && (amount<0x40) && ((amount+i)<(bmp_head->width*bmp_head->height)))
				{
					amount++;
				}
				printf("0x%04x, ", ((amount-1)<<10) + buffer[i]);
				if(!((column+1)%12))
					printf("\n");
				column++;
				i+=(amount-1);
				compressedtargetsize+=2;
			}
		}
		printf("\n};\n");
		fprintf(stderr, "Final size: %u\n", compressedtargetsize);
		fprintf(stderr, "Compressionrate: %.2f\n", 1.0-((double)compressedtargetsize)/((double)targetsize));
	}
	else
	{
		fprintf(stderr, "RAW mode\n");
		for(i=0; i<(bmp_head->width*bmp_head->height); i++)
		{
			if(targetcolordepth == 16)
			{
				printf("0x%04x, ", buffer[i]);
				if(!((i+1)%12))
					printf("\n");
			}
			else if(targetcolordepth == 8)
			{
				printf("0x%02x, ", buffer8[i]);
				if(!((i+1)%16))
					printf("\n");
			}
		}
		printf("\n};\n");
		fprintf(stderr, "Final size = Target Size\n");
	}
	




	if(targetcolordepth == 16)
		free(buffer);
	else if(targetcolordepth == 8)
		free(buffer8);
	fclose(fd_in);
	return(0);
}
